/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.layout;

import static java.lang.Math.max;
import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_CONNECTOR_SIZE;
import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_GRID_SIZE;
import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_SHAPE_MINIMUM_WIDTH;
import static org.fortiss.tooling.base.utils.LayoutDataUtils.getNodeSize;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.pickInstanceOf;

import org.fortiss.tooling.base.model.base.EntryConnectorBase;
import org.fortiss.tooling.base.model.element.IConnection;
import org.fortiss.tooling.base.model.element.IConnector;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.base.model.layout.Dimension;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.kernel.model.INamedElement;

import javafx.geometry.Bounds;
import javafx.scene.text.Text;

/**
 * {@link IAutoLayouterTapeMeasure} base class of models to be displayed in JavaFX-based editors.
 * 
 * @author barner
 */
public abstract class DiagramTapeMeasure implements IAutoLayouterTapeMeasure {
	/**
	 * Returns the an estimation of the text extent of the given {@code element}'s label (may be
	 * {@code null}).
	 */
	protected Bounds getTextExtent(IModelElement element) {
		if(!(element instanceof INamedElement)) {
			return null;
		}
		String name = ((INamedElement)element).getName();
		Text text = new Text(name);
		return text.getBoundsInLocal();
	}

	/**
	 * Returns the maximum of the given length and a default length derived from the element's
	 * connector count.
	 */
	protected int getElementLength(int length, IHierarchicElement element) {
		int numElemConnectors = element.getConnectors().size();
		int numElemInputConnectors =
				pickInstanceOf(EntryConnectorBase.class, element.getConnectors()).size();
		int numConnectors = max(numElemInputConnectors, numElemConnectors - numElemInputConnectors);

		return max((1 + 2 * numConnectors) * DEFAULT_CONNECTOR_SIZE, length);
	}

	/** {@inheritDoc} */
	@Override
	public int getElementWidth(IHierarchicElement element) {
		Dimension dimension = getNodeSize((ILayoutedModelElement)element);

		int width = max(DEFAULT_SHAPE_MINIMUM_WIDTH, dimension.getWidth());
		Bounds textExtent = getTextExtent(element);
		if(textExtent != null) {
			width = max(width, (int)textExtent.getWidth() + 6 * DEFAULT_GRID_SIZE);
		}

		return width;
	}

	/** {@inheritDoc} */
	@Override
	public int getElementHeight(IHierarchicElement element) {
		Dimension dimension = getNodeSize((ILayoutedModelElement)element);
		return dimension.getHeight();
	}

	/** {@inheritDoc} */
	@Override
	public int getConnectorWidth(IConnector connector) {
		return DEFAULT_CONNECTOR_SIZE;
	}

	/** {@inheritDoc} */
	@Override
	public int getConnectorHeight(IConnector connector) {
		return DEFAULT_CONNECTOR_SIZE;
	}

	/** {@inheritDoc} */
	@Override
	public int getConnectionWidth(IConnection connection) {
		Bounds textExtend = getTextExtent(connection);
		return textExtend != null ? (int)textExtend.getWidth() : 0;
	}

	/** {@inheritDoc} */
	@Override
	public int getConnectorLabelWidth(IConnector connector) {
		Bounds textExtend = getTextExtent(connector);
		return textExtend != null ? (int)textExtend.getWidth() : 0;
	}
}
