/*-------------------------------------------------------------------------+
| Copyright 2023 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.quality;

import org.fortiss.tooling.base.model.base.EntryConnectorBase;
import org.fortiss.tooling.base.model.base.ExitConnectorBase;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.ext.quality.data.MetricKey;
import org.fortiss.tooling.ext.quality.data.MetricTreeNode;
import org.fortiss.tooling.kernel.model.IIdLabeled;
import org.fortiss.tooling.kernel.model.INamedCommentedElement;
import org.fortiss.tooling.kernel.utils.EcoreUtils;

/**
 * {@link IMetricProvider} to count the ratio of filled out comments of
 * INamedCommentedElements.
 * 
 * @author blaschke
 * @author groh
 */
public class HierarchicElementProvider implements IMetricProvider<IHierarchicElement> {
	/**
	 * returns an array of integers or <String,Double> Map Number of children of the
	 * hierarchical Element (on this level) Number of ports on the hierarchical
	 * Element (on this level) Number of channels in the hierarchical element (on
	 * this level) Number of totalPorts in whole tree Number of totalElements in
	 * whole tree Average SubTree Size
	 */

	/**
	 * List containing all HierarchicElementSizeProvider for components inside the
	 * IHierarchicElement
	 */

	/** */
	public HierarchicElementProvider() {
		super();
	}

	/** {@inheritDoc} */
	@Override
	public void collectMetrics(MetricTreeNode node, IHierarchicElement currentElement) {

		var metrics = node.getStoredMetrics();
		if(node.getName() == null) {
			boolean isElementCommentable = currentElement instanceof INamedCommentedElement;
			metrics.put(MetricKey.NUMBER_OF_TOTAL_COMMENTABLE_ELEMENTS,
					isElementCommentable ? 1.0 : 0.0);
			String comment = isElementCommentable
					? ((INamedCommentedElement)currentElement).getComment() : null;
			metrics.put(MetricKey.NUMBER_OF_TOTAL_COMMENTED_ELEMENTS,
					comment != null && comment != "" ? 1.0 : 0.0);
			if(isElementCommentable) {
				node.setName(((INamedCommentedElement)currentElement).getName());
			} else {
				node.setName("Unnamable Element");
			}

			metrics.put(MetricKey.UNQIUE_ID, (currentElement instanceof IIdLabeled)
					? ((IIdLabeled)currentElement).getId() : -1.0);
		}

		var connectors = currentElement.getConnectors();
		metrics.put(MetricKey.NUMBER_OF_CONNECTORS, (double)connectors.size());

		metrics.put(MetricKey.NUMBER_OF_CONTAINED_ELEMENTS,
				(double)currentElement.getContainedElements().size());
		metrics.put(MetricKey.NUMBER_OF_CHANNELS, (double)currentElement.getConnections().size());
		// depth metrics
		metrics.put(MetricKey.NUMBER_OF_TOTAL_CONNECTORS, (double)connectors.size());
		var entry_connectors = EcoreUtils.pickInstanceOf(EntryConnectorBase.class, connectors);
		var exit_connectors = EcoreUtils.pickInstanceOf(ExitConnectorBase.class, connectors);
		metrics.put(MetricKey.NUMBER_OF_TOTAL_ENTRY_CONNECTORS, (double)entry_connectors.size());
		metrics.put(MetricKey.NUMBER_OF_TOTAL_EXIT_CONNECTORS, (double)exit_connectors.size());

		metrics.put(MetricKey.NUMBER_OF_TOTAL_ELEMENTS, 1.0);

		metrics.put(MetricKey.NUMBER_OF_TOTAL_LEAF_ELEMENTS,
				currentElement.getContainedElements().size() == 0 ? 1.0 : 0.0);
	}
}
