/*-------------------------------------------------------------------------+
| Copyright 2023 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.quality.data;

import java.util.ArrayList;
import java.util.List;

/**
 * 
 * @author groh
 */
public enum MetricKey {

	//
	// Metrics provided by the HierarchicElementProvider
	//
	/** Unique ID of the element */
	UNQIUE_ID(false),

	/** Key for the metric counting ports */
	NUMBER_OF_CONNECTORS(false),
	/** Key for the metric counting elements */
	NUMBER_OF_CONTAINED_ELEMENTS(false),
	/** Key for the metric counting channels */
	NUMBER_OF_CHANNELS(false),
	/** Safety level of the element */
	SAFETY_LEVEL(false, true),

	//
	// Metrics which are collected over all children nodes
	//
	/** Key for the metric counting the total amount of ports contained */
	NUMBER_OF_TOTAL_CONNECTORS(false),
	/** Key for the metric counting the total amount of input ports contained */
	NUMBER_OF_TOTAL_ENTRY_CONNECTORS(false),
	/** Key for the metric counting the total amount of output ports contained */
	NUMBER_OF_TOTAL_EXIT_CONNECTORS(false),
	/** Key for the metric counting the total amount of elements contained */
	NUMBER_OF_TOTAL_ELEMENTS(false),
	/** Key for the metric counting the total amount of elements which can be commented */
	NUMBER_OF_TOTAL_COMMENTABLE_ELEMENTS(false),
	/** Key for the metric counting the total amount of elements which are commented */
	NUMBER_OF_TOTAL_COMMENTED_ELEMENTS(false),
	/**
	 * Key for the metric counting the total amount of elements which do not contain other elements
	 */
	NUMBER_OF_TOTAL_LEAF_ELEMENTS(true),
	/**
	 * Key for the metric counting of the total amount of constant values in all code specifications
	 */
	NUMBER_OF_TOTAL_CODE_CONSTANTS(true),

	/** Cyclomatic complexity number as sum over all components */
	NUMBER_OF_TOTAL_CODE_CYCLOMATIC_COMPLEXITY(true);

	/** Metrics which are a combination from the data of multiple nodes */
	private static List<MetricKey> collectorKeys = new ArrayList<>();

	/**
	 * Also see {@link MetricKey#collectorKeys}
	 * 
	 * @return the keys which are combined from all children of a node
	 */
	public static List<MetricKey> getCollectorKeys() {
		return collectorKeys;
	}

	/** Used for populating the {@link MetricKey#getCollectorKeys()} Method */
	private boolean isCollectorKey;
	/** Determines if the corresponding value to this key is a string or not */
	private boolean isStringValue;

	/**
	 * Basic constructor
	 * 
	 * @param isCollectorKey
	 *            decides if this is a key returned by {@link MetricKey#getCollectorKeys()}
	 */
	private MetricKey(boolean isCollectorKey) {
		this(isCollectorKey, false);
	}

	/**
	 * @param isCollectorKey
	 *            decides if this is a key returned by {@link MetricKey#getCollectorKeys()}
	 * @param isStringValue
	 *            marks if the value associated with this key is a string
	 */
	private MetricKey(boolean isCollectorKey, boolean isStringValue) {
		this.isCollectorKey = isCollectorKey;
		this.isStringValue = isStringValue;
	}

	/**
	 * If this value is true, it should be expected that the value corresponding to this key should
	 * be a string and be stored in an appropriate place
	 * 
	 * @return if the value associated with this key is a string
	 */
	public boolean isStringValue() {
		return isStringValue;
	}

	static {
		for(MetricKey key : MetricKey.values()) {
			if(key.isCollectorKey) {
				collectorKeys.add(key);
			}
		}
	}
}
