/*-------------------------------------------------------------------------+
| Copyright 2023 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.quality.data;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.function.Consumer;

import org.fortiss.tooling.ext.quality.IMetricProvider;

/**
 * A general class for storing all kinds of metric data in a hierarchical tree structure.
 * 
 * Allows for efficient storage and retrieval of data
 * 
 * @author groh
 */
public class MetricTreeNode {

	/** stores the children of this tree node */
	private List<MetricTreeNode> children;

	/**
	 * This map stores a relationship between a {@link IMetricProvider} and
	 * {@link IMetricDataContainer}.
	 * 
	 * This allows all {@linkplain IMetricProvider} to deposit data inside this structure, and
	 * quickly retrieve it for processing
	 * 
	 * The stored data can then also be accessed by other methods, for displaying information,
	 * creating statistics, etc.
	 */
	private Map<Class<? extends IMetricProvider<?>>, IMetricDataContainer> dataContainers;

	/** Map containing all metrics which are a double */
	private Map<MetricKey, Double> storedMetrics;

	/** Map containing all metrics which are a double */
	private Map<MetricKey, String> storedStrings;

	/** name of the element which is associated with this data */
	private String name;

	/**
	 * Constructs a new node
	 */
	public MetricTreeNode() {
		children = new ArrayList<>();
		dataContainers = new HashMap<>();
		storedMetrics = new HashMap<>();
		storedStrings = new HashMap<>();
	}

	/**
	 * @return the children of this node
	 */
	public List<MetricTreeNode> getChildren() {
		return children;
	}

	/**
	 * See {@link MetricTreeNode#dataContainers}
	 * 
	 * @return the map storing all DataContainers in this node
	 */
	public Map<Class<? extends IMetricProvider<?>>, IMetricDataContainer> getDataContainers() {
		return dataContainers;
	}

	/**
	 * @return the name of the element which is associated with this data
	 */
	public String getName() {
		return name;
	}

	/** Set the name */
	public void setName(String name) {
		this.name = name;
	}

	/**
	 * @return the stored metrics map
	 */
	public Map<MetricKey, Double> getStoredMetrics() {
		return storedMetrics;
	}

	/**
	 * @return the stored strings map
	 */
	public Map<MetricKey, String> getStoredStrings() {
		return storedStrings;
	}

	/**
	 * Traverses the tree in post-order
	 * 
	 * @param consumer
	 *            which is applied to each node
	 */
	public void traverseTree(Consumer<MetricTreeNode> consumer) {

		// traverse all children
		for(var child : getChildren()) {
			child.traverseTree(consumer);
		}

		// apply consumer to self
		consumer.accept(this);
	}
}
