/*-------------------------------------------------------------------------+
| Copyright 2013 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.quality.service;

import static java.util.Collections.emptyList;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.ext.quality.IMetricProvider;
import org.fortiss.tooling.ext.quality.data.MetricDataManager;
import org.fortiss.tooling.ext.quality.data.MetricKey;
import org.fortiss.tooling.ext.quality.data.MetricTreeNode;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.introspection.IIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectionItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectiveKernelService;
import org.fortiss.tooling.kernel.model.IProjectRootElement;
import org.fortiss.tooling.kernel.service.IKernelIntrospectionSystemService;
import org.fortiss.tooling.kernel.service.IMigrationService;
import org.fortiss.tooling.kernel.service.base.EObjectAwareServiceBase;
import org.fortiss.tooling.kernel.utils.EcoreUtils;

/**
 * This class implements the {@link IMigrationService} interface.
 * 
 * @author blaschke
 */
public class ModelQualityService extends EObjectAwareServiceBase<IMetricProvider<?>>
		implements IIntrospectiveKernelService, IModelQualityService {

	/** The singleton instance. */
	private static final ModelQualityService INSTANCE = new ModelQualityService();

	/** Returns singleton instance of the service. */
	public static ModelQualityService getInstance() {
		return INSTANCE;
	}

	/** The connector extension point ID. */
	private static final String EXTENSION_POINT_NAME =
			"org.fortiss.tooling.ext.quality.metricService";

	/** The connector configuration element name. */
	private static final String CONFIGURATION_ELEMENT_NAME = "metricService";

	/** The connector attribute name. */
	private static final String ATTRIBUTE_NAME = "metricService";

	/** Static instance of the MetricDataManager */
	public static MetricDataManager metricDataManagerInstance = new MetricDataManager();

	/** {@inheritDoc} */
	@Override
	public void startService() {
		IKernelIntrospectionSystemService.getInstance().registerService(this);
	}

	/** Registers the migration provider with the service. */
	@Override
	public <T extends EObject> void registerMetricProvider(IMetricProvider<T> provider,
			Class<T> modelElementClass) {
		addHandler(modelElementClass, provider);
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionDescription() {
		return getIntrospectionLabel() +
				"\n\nThis service allows the tracking of metrical development in modeling." +
				"\n the service outputs the current metric analysis of the model and stores" +
				"\n the data within logs for later history analysis." +
				"\n\nThe service extension point is '" + EXTENSION_POINT_NAME + "'.";
	}

	/**
	 * @param topLvl
	 *            the file project on which the analysis should be performed
	 */
	@Override
	public void performMetricAnalysis(ITopLevelElement topLvl) {

		for(IProjectRootElement rootElement : EcoreUtils
				.getChildrenWithType(topLvl.getRootModelElement(), IProjectRootElement.class)) {
			if(rootElement instanceof IHierarchicElement) {

				EList<IHierarchicElement> elements =
						((IHierarchicElement)rootElement).getContainedElements();
				if(elements.isEmpty()) {
					continue;
				}
				var root_nodes = metricDataManagerInstance.getRootNodes();

				// Get first element, there should not be any other elements
				IHierarchicElement firstElement = elements.get(0);
				MetricTreeNode root_node = new MetricTreeNode();
				recursivlyCollectMetrics(root_node, firstElement);
				root_nodes.put(rootElement, root_node);
			}
		}
		// Store all currently saved metrics to the csv
		metricExtractionToCSV(metricDataManagerInstance.getRootNodes());
	}

	/**
	 * @param node
	 *            to which the data is collected
	 * @param currentElement
	 *            element to collect the data from
	 */
	private void recursivlyCollectMetrics(MetricTreeNode node, IHierarchicElement currentElement) {

		var manager = metricDataManagerInstance;

		collectMetrics(node, currentElement);

		// Check if any of the specifications is an IHierarchicElement
		// This is for example the case with an StateAutomaton
		Optional<IHierarchicElement> hierarchicElementOptional = currentElement.getSpecifications()
				.stream().filter(s -> s instanceof IHierarchicElement)
				.map(s -> (IHierarchicElement)s).findAny();
		if(hierarchicElementOptional.isPresent()) {
			// Do not collect statistics for the current element, but instead collect them for the
			// specification
			IHierarchicElement specificationElement = hierarchicElementOptional.get();
			EList<IHierarchicElement> containedElements =
					specificationElement.getContainedElements();
			if(containedElements.size() == 1) {
				// Add reference from the element to this, so the lookups works as expected
				manager.getHierarchicLookupTable().put(specificationElement, node);
				// Skip the specification element to get a more useful tree structure
				recursivlyCollectMetrics(node, containedElements.get(0));
			} else {
				recursivlyCollectMetrics(node, specificationElement);
			}
			// Add reference from the element to this, so the lookups works as expected
			manager.getHierarchicLookupTable().put(currentElement, node);
		} else {

			// Iterate over all children and
			for(IHierarchicElement containedElement : currentElement.getContainedElements()) {
				MetricTreeNode child = new MetricTreeNode();
				node.getChildren().add(child);
				recursivlyCollectMetrics(child, containedElement);

				for(MetricKey key : MetricKey.getCollectorKeys()) {
					var childMetrics = child.getStoredMetrics();
					if(childMetrics.containsKey(key)) {
						node.getStoredMetrics().merge(key, child.getStoredMetrics().get(key),
								Double::sum);
					}
				}
			}

			manager.getHierarchicLookupTable().put(currentElement, node);
		}
	}

	/**
	 * @param node
	 *            to which the data is collected
	 * @param element
	 *            element to collect the data from
	 */
	@SuppressWarnings("unchecked")
	private void collectMetrics(MetricTreeNode node, IHierarchicElement element) {

		List<IMetricProvider<?>> providers = getAllMetricProviders(element);

		if(providers.isEmpty()) {
			// it should not happen that no provider is found
			throw new IllegalStateException();
		}

		for(IMetricProvider<?> provider : providers) {
			((IMetricProvider<EObject>)provider).collectMetrics(node, element);
		}
	}

	/** Get all suitable {@link IMetricProvider} for the given input. */
	private List<IMetricProvider<?>> getAllMetricProviders(EObject input) {
		List<IMetricProvider<?>> providers = new ArrayList<>();
		for(Entry<Class<?>, List<IMetricProvider<?>>> migEntry : handlerMap.entrySet()) {
			if(migEntry.getKey().isAssignableFrom(input.getClass())) {
				providers.addAll(migEntry.getValue());
			}
		}
		return providers;
	}

	/** {@inheritDoc} */
	@Override
	protected String getExtensionPointName() {
		return EXTENSION_POINT_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected String getConfigurationElementName() {
		return CONFIGURATION_ELEMENT_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected String getHandlerClassAttribute() {
		return ATTRIBUTE_NAME;
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionLabel() {
		return "Metric Service";
	}

	/** {@inheritDoc} */
	@Override
	public List<IIntrospectionItem> getIntrospectionItems() {
		return emptyList();
	}

	/** {@inheritDoc} */
	@Override
	public boolean showInIntrospectionNavigation() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public IIntrospectionDetailsItem getDetailsItem() {
		return null;
	}

	/**
	 * Method to write the data into a csv
	 * 
	 * @param map
	 *            to convert into csv
	 * 
	 */
	public void metricExtractionToCSV(Map<IProjectRootElement, MetricTreeNode> map) {

		// check if there is any data to export
		if(map == null || map.isEmpty()) {
			return;
		}

		// path to csv file
		Path path = Paths.get("data_mega_cool.csv");

		List<String> allKeys = new ArrayList<>();
		boolean createNewIndex = true;

		if(Files.exists(path)) {
			// Read first line and try to extract the keys from the already existing file
			try(BufferedReader reader = new BufferedReader(new FileReader(path.toFile()))) {
				String line = reader.readLine();
				if(line != null) {
					allKeys = Arrays.asList(line.split(","));
					createNewIndex = false;
				}
			} catch(IOException e) {
				e.printStackTrace();
			}
		}

		try(var writer = new BufferedWriter(new FileWriter(path.toFile(), true))) {

			if(createNewIndex) {
				// Create new index and write it into the first line of the file
				allKeys.add("timestamp");
				allKeys.add("root_name");
				allKeys.add("name");
				allKeys.add("children");
				// Collect all MetricKeys, convert them to a string and append them to the allKeys
				// list
				allKeys.addAll(Arrays.stream(MetricKey.values()).map(MetricKey::toString)
						.map(String::toLowerCase).collect(Collectors.toList()));

				// Write first line
				writer.write(String.join(",", allKeys));
				writer.write("\n");
			}
			// Remove first 4 keys as they are not provided by the provider
			List<String> valueKeys = allKeys.subList(4, allKeys.size());

			LocalDateTime dateTime = LocalDateTime.now();
			DateTimeFormatter formatter = DateTimeFormatter.ISO_LOCAL_DATE_TIME;
			String formattedDateTime = dateTime.format(formatter);

			map.forEach((rootProject, rootNode) -> {
				// iterate over all project roots
				String rootName = rootProject.getName();

				rootNode.traverseTree(node -> {
					var metrics = node.getStoredMetrics();
					var strings = node.getStoredStrings();
					// collect IDs of children
					String children_ids = node.getChildren().stream()
							.map(n -> n.getStoredMetrics().get(MetricKey.UNQIUE_ID))
							.map(id -> id.toString()).reduce((a, b) -> a + " " + b).orElse("");

					// These 4 data entries are just concatenated in front
					var startStream =
							Stream.of(formattedDateTime, rootName, node.getName(), children_ids);
					// Collect all values from provider in the correct order and combine
					var values = Stream.concat(startStream,
							valueKeys.stream().map(String::toUpperCase).map(MetricKey::valueOf)
									.map(k -> k.isStringValue() ? strings.get(k) : metrics.get(k))
									.map(String::valueOf))
							.collect(Collectors.joining(","));

					try {
						// write values
						writer.write(values);
						writer.write("\n");
					} catch(IOException e) {
						e.printStackTrace();
					}
				});
			});
		} catch(IOException e) {
			e.printStackTrace();
		}
	}
}
