/*-------------------------------------------------------------------------+
| Copyright 2021 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.reuse.ui.view.fx;

import static org.conqat.ide.commons.ui.dialog.MessageUtils.confirm;
import static org.eclipse.jface.window.Window.OK;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsBasics.getAllLocalReuseLibraries;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsBasics.getFirstReuseElementUUID;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsBasics.getFirstSourceLibraryOfElement;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsBasics.getLibraryIDName;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsBasics.getReuseElementComment;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsBasics.getReuseElementLastUpdateString;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsBasics.getReuseElementName;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsManipulation.deleteReuseElementInLibraryFile;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsManipulation.deleteReuseLibrary;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsManipulation.updateReuseElementFileWithComment;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsManipulation.updateReuseElementFileWithName;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.jface.dialogs.IInputValidator;
import org.eclipse.jface.dialogs.InputDialog;
import org.eclipse.swt.widgets.Display;
import org.fortiss.tooling.common.ui.javafx.control.treetableview.DynamicTreeTableUIProviderBase;
import org.fortiss.tooling.ext.reuse.model.ReuseLibrary;
import org.fortiss.tooling.kernel.ui.service.IModelElementHandlerService;

import javafx.scene.Node;
import javafx.scene.control.ContextMenu;
import javafx.scene.control.MenuItem;

/**
 * UIProvider for the FX-based {@link ReuseLibrary} tree table.
 *
 * @author bergemann
 */
final class ReuseLibraryFXTreeTableUIProvider extends DynamicTreeTableUIProviderBase<EObject> {

	/** The {@link ReuseLibraryFXController} this UI provider was created for. */
	private final ReuseLibraryFXController viewController;

	/** All names that are already taken when using the name validator. */
	private List<String> blockedNames;

	/**
	 * Constructor.
	 *
	 * @param reuseLibraryViewFXController The used FX controller for the tree table
	 *                                     of this UI provider.
	 */
	ReuseLibraryFXTreeTableUIProvider(ReuseLibraryFXController reuseLibraryViewFXController) {
		this.viewController = reuseLibraryViewFXController;
		blockedNames = new ArrayList<String>();
	}

	/** {@inheritDoc} */
	@Override
	public String getLabel(EObject element, int column) {
		switch (column) {
		case 0:
			// Library Name (only if it is a ReuseLibrary)
			if (element instanceof ReuseLibrary) {
				return getReuseElementName(element);
			} else {
				return "";
			}
		case 1:
			// Element Name (only if it is NOT a ReuseLibrary)
			if (element instanceof ReuseLibrary) {
				return "";
			} else {
				return getReuseElementName(element);
			}
		case 2:
			// UUID
			return getFirstReuseElementUUID(element);
		case 3:
			// Last Update
			boolean stringIsForDisplay = true;
			return getReuseElementLastUpdateString(element, stringIsForDisplay);
		case 4:
			// Comment
			return getReuseElementComment(element);
		default:
			return "X";
		}
	}

	/** {@inheritDoc} */
	@Override
	public Node getIconNode(EObject element, int column) {
		if (isElementMatchedWithItsColumn(element, column)) {
			return IModelElementHandlerService.getInstance().getFXIcon(element);
		}
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public boolean isEditable(int column) {
		// allow (re)commenting
		return column == 4;
	}

	/** {@inheritDoc} */
	@Override
	public void updateValue(EObject element, int column, Object value) {
		// updating column 4 is (re)commenting
		if (column == 4 && value instanceof String) {
			updateReuseElementFileWithComment(element, (String) value);
			viewController.updateReuseLibraryView();
		}
	}

	/** {@inheritDoc} */
	@Override
	public ContextMenu createContextMenu(EObject element, int column) {
		if (isElementMatchedWithItsColumn(element, column)) {
			ContextMenu contextMenu = new ContextMenu();
			contextMenu.getItems().add(createDeleteMenuForReuse(element));
			contextMenu.getItems().add(createRenameMenuForReuse(element));
			return contextMenu;
		}
		return null;
	}

	/** Creates a delete menu entry for the given {@link ReuseLibrary}. */
	private MenuItem createDeleteMenuForReuse(EObject target) {
		MenuItem deleteMenuItem = null;

		if (target instanceof ReuseLibrary) {
			deleteMenuItem = new MenuItem("Delete Library");
			ReuseLibrary targetLibrary = (ReuseLibrary) target;
			deleteMenuItem.setOnAction(event -> {

				String dialogTitle = "Confirm Deleting A Reuse Library";
				String confirmationMessage = "Do you want to completely delete the following reuse library?"
						+ "\n\nDelete: " + getLibraryIDName(targetLibrary);
				boolean confirmed = confirm(dialogTitle, confirmationMessage);

				if (confirmed) {
					deleteReuseLibrary(targetLibrary);
					viewController.updateReuseLibraryView();
					// TODO #4151 (feature): add PersistentService for undo, etc. and then use here
					// runAsCommand with context for the deletion
				}
			});
		} else {
			deleteMenuItem = new MenuItem("Delete Reuse Element");
			deleteMenuItem.setOnAction(event -> {

				ReuseLibrary sourceLibrary = getFirstSourceLibraryOfElement(target);
				String dialogTitle = "Confirm Deleting A Reuse Element";
				String confirmationMessage = "Do you want to completely delete the following reuse element?"
						+ "\n\nDelete: " + getReuseElementName(target) + "\nFrom library: "
						+ getLibraryIDName(sourceLibrary);
				boolean confirmed = confirm(dialogTitle, confirmationMessage);

				if (confirmed) {
					deleteReuseElementInLibraryFile(target);
					viewController.updateReuseLibraryView();
					// TODO #4151 (feature): add PersistentService for undo, etc. and then use here
					// runAsCommand with context for the deletion
				}
			});
		}

		return deleteMenuItem;
	}

	/** Creates a rename menu entry for the given {@link ReuseLibrary}. */
	private MenuItem createRenameMenuForReuse(EObject target) {
		MenuItem renameMenuItem = null;
		String type = "";
		List<String> blockedNamesTmp = new ArrayList<String>();

		if (target instanceof ReuseLibrary) {
			type = "Library";
			for (ReuseLibrary library : getAllLocalReuseLibraries()) {
				blockedNamesTmp.add(getReuseElementName(library));
			}
		} else {
			type = "Reuse Element";
			ReuseLibrary targetLibrary = getFirstSourceLibraryOfElement(target);
			for (EObject libraryElement : targetLibrary.getReuseElementList()) {
				blockedNamesTmp.add(getReuseElementName(libraryElement));
			}
		}

		final String actionName = "Rename " + type;
		final String requestText = "Please enter a new name for the selected " + type.toLowerCase();
		final String defaultName = getReuseElementName(target);
		renameMenuItem = new MenuItem(actionName);
		renameMenuItem.setOnAction(event -> {
			String newName = requestName(actionName, requestText, defaultName, blockedNamesTmp);
			if (newName != null) {
				updateReuseElementFileWithName(target, newName);
				viewController.updateReuseLibraryView();
			}
		});

		return renameMenuItem;
	}

	/**
	 * Returns a {@link String} that was given by the user when asked for a new name
	 * of the given element. Returns null if the user canceled the action.
	 */
	private String requestName(String shellTitle, String requestText, String defaultName,
			List<String> blockedNamesTmp) {
		blockedNames.clear();
		blockedNames = blockedNamesTmp;

		final InputDialog creationDialog = new InputDialog(Display.getCurrent().getActiveShell(), shellTitle,
				requestText + "\n ", defaultName, new NewNameValidator());
		if (creationDialog.open() != OK) {
			return null;
		}
		return creationDialog.getValue();
	}

	/**
	 * Checks if the given {@link EObject} matches with the given column.
	 *
	 * For example: a reuse library should be in the first column and everything
	 * else (as reuse element) in the second.
	 */
	private boolean isElementMatchedWithItsColumn(EObject element, int column) {
		if (element != null) {
			if ((column == 0 && element instanceof ReuseLibrary)
					|| (column == 1 && !(element instanceof ReuseLibrary))) {
				return true;
			}
		}
		return false;
	}

	/** Validator for the new name. */
	private class NewNameValidator implements IInputValidator {

		/** {@inheritDoc} */
		@Override
		public String isValid(String newText) {
			if (newText.equals("")) {
				return "A name needs to be specified.";
			}
			for (String name : blockedNames) {
				if (newText.equals(name)) {
					return "This name already exists. Please use another.";
				}
			}
			return null;
		}
	}
}
